;;;; Syntax Checking
;;; written by Alan Bawden
;;; modified by Chris Hanson

(define (syntax-check pattern form)
  (if (not (syntax-match? (cdr pattern) (cdr form)))
      (syntax-error "ill-formed special form" form)))

(define (syntax-match? pattern object)
  (define (check-llist object)
    (let loop ((seen '()) (object object))
      (or (null? object)
	  (if (identifier? object)
	      (not (memq object seen))
	      (and (pair? object)
		   (identifier? (car object))
		   (not (memq (car object) seen))
		   (loop (cons (car object) seen) (cdr object)))))))
  (let ((match-error
	 (lambda ()
	   (error "ill-formed pattern" pattern))))
    (cond ((symbol? pattern)
	   (case pattern
	     ((identifier) (identifier? object))
	     ((datum expression form) #t)
	     ((r4rs-bvl) (check-llist object))
	     ((dsssl-bvl) (or (##sys#extended-lambda-list? object) (check-llist object)))
	     ;;((mit-bvl) (lambda-list? object))
	     (else (match-error))))
	  ((pair? pattern)
	   (case (car pattern)
	     ((*)
	      (if (pair? (cdr pattern))
		  (let ((head (cadr pattern))
			(tail (cddr pattern)))
		    (let loop ((object object))
		      (or (and (pair? object)
			       (syntax-match? head (car object))
			       (loop (cdr object)))
			  (syntax-match? tail object))))
		  (match-error)))
	     ((+)
	      (if (pair? (cdr pattern))
		  (let ((head (cadr pattern))
			(tail (cddr pattern)))
		    (and (pair? object)
			 (syntax-match? head (car object))
			 (let loop ((object (cdr object)))
			   (or (and (pair? object)
				    (syntax-match? head (car object))
				    (loop (cdr object)))
			       (syntax-match? tail object)))))
		  (match-error)))
	     ((?)
	      (if (pair? (cdr pattern))
		  (or (and (pair? object)
			   (syntax-match? (cadr pattern) (car object))
			   (syntax-match? (cddr pattern) (cdr object)))
		      (syntax-match? (cddr pattern) object))
		  (match-error)))
	     ((quote)
	      (if (and (pair? (cdr pattern))
		       (null? (cddr pattern)))
		  (eqv? (cadr pattern) object)
		  (match-error)))
	     (else
	      (and (pair? object)
		   (syntax-match? (car pattern) (car object))
		   (syntax-match? (cdr pattern) (cdr object))))))
	  (else
	   (eqv? pattern object)))))
